package cc.fascinated.bat.features.birthday.profile;

import cc.fascinated.bat.common.Profile;
import cc.fascinated.bat.model.BatGuild;
import lombok.Getter;
import lombok.Setter;
import net.dv8tion.jda.api.entities.Guild;
import net.dv8tion.jda.api.entities.Member;
import net.dv8tion.jda.api.entities.User;
import net.dv8tion.jda.api.entities.channel.concrete.TextChannel;

import java.util.*;

/**
 * @author Fascinated (fascinated7)
 */
@Getter
@Setter
public class BirthdayProfile extends Profile {
    private static final String DEFAULT_MESSAGE = "Happy Birthday {user} :tada: :birthday: You are now {age} years old!";

    /**
     * The list of birthdays that are being tracked
     */
    private Map<String, Date> birthdays;

    /**
     * The channel ID of the birthday feed
     */
    private String channelId;

    /**
     * The message to send when it is a user's birthday
     */
    private String message = DEFAULT_MESSAGE;

    public BirthdayProfile() {
        super("birthday");
    }

    /**
     * Adds a birthday to be tracked
     *
     * @param userId   the id of the user to track
     * @param birthday the birthday of the user
     */
    public void addBirthday(String userId, Date birthday) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }
        birthdays.put(userId, birthday);
    }

    /**
     * Removes a birthday from being tracked
     *
     * @param userId the id of the user to remove
     */
    public void removeBirthday(String userId) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }
        birthdays.remove(userId);
    }

    /**
     * Gets the birthday of a user
     *
     * @param userId the id of the user
     * @return the birthday of the user
     */
    public Date getBirthday(String userId) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }
        return birthdays.get(userId);
    }

    /**
     * Checks if the profile has a channel setup
     *
     * @return if the profile has a channel setup
     */
    public boolean hasChannelSetup() {
        return channelId != null;
    }

    /**
     * Calculates the age of a user
     *
     * @param userId the id of the user
     * @return the age of the user
     */
    public int calculateAge(String userId) {
        Date birthday = getBirthday(userId);
        if (birthday == null) {
            return 0; // or throw an exception
        }

        Calendar birthdayCalendar = Calendar.getInstance();
        birthdayCalendar.setTime(birthday);

        Calendar today = Calendar.getInstance();

        int age = today.get(Calendar.YEAR) - birthdayCalendar.get(Calendar.YEAR);

        // Check if the birthday hasn't occurred yet this year
        if (today.get(Calendar.DAY_OF_YEAR) < birthdayCalendar.get(Calendar.DAY_OF_YEAR)) {
            age--;
        }

        return age;
    }

    /**
     * Validates the profiles configuration
     *
     * @param guild the guild to validate
     * @return if the profile is valid
     */
    public boolean validate(BatGuild guild) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }

        List<String> toRemove = new ArrayList<>();
        Guild discordGuild = guild.getDiscordGuild();
        for (Map.Entry<String, Date> entry : birthdays.entrySet()) {
            String userId = entry.getKey();
            Date birthday = entry.getValue();

            if (userId == null || birthday == null) { // this should never happen
                continue;
            }

            // Check if the user is still in the guild, if not remove them
            Member member = discordGuild.getMemberById(userId);
            if (member == null) {
                toRemove.add(userId);
            }
        }

        for (String userId : toRemove) {
            birthdays.remove(userId);
        }

        if (channelId == null) {
            return false;
        }

        if (discordGuild.getTextChannelById(channelId) == null) {
            channelId = null;
            return false;
        }
        return true;
    }

    /**
     * Checks to see if it is any user's birthday
     *
     * @param guild the guild to check for
     */
    public void checkBirthdays(BatGuild guild) {
        if (!validate(guild)) {
            return;
        }

        // Get today's day and month
        Calendar today = Calendar.getInstance();
        int todayDay = today.get(Calendar.DAY_OF_MONTH);
        int todayMonth = today.get(Calendar.MONTH); // Note: January is 0

        Iterator<Map.Entry<String, Date>> iterator = birthdays.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<String, Date> entry = iterator.next();
            Date birthday = entry.getValue();

            if (birthday == null) {
                iterator.remove();
                continue;
            }

            Calendar birthdayCalendar = Calendar.getInstance();
            birthdayCalendar.setTime(birthday);

            int birthdayDay = birthdayCalendar.get(Calendar.DAY_OF_MONTH);
            int birthdayMonth = birthdayCalendar.get(Calendar.MONTH);

            // Check if the birthday matches today's day and month
            if (birthdayDay == todayDay && birthdayMonth == todayMonth) {
                notifyBirthday(guild, entry.getKey());
            }
        }
    }

    /**
     * Notifies a user that it is their birthday
     *
     * @param guild  the guild
     * @param userId the id of the user
     */
    private void notifyBirthday(BatGuild guild, String userId) {
        Guild discordGuild = guild.getDiscordGuild();
        Member member = discordGuild.getMemberById(userId);
        if (member == null) {
            return;
        }

        TextChannel channel = discordGuild.getTextChannelById(channelId);
        if (channel == null) { // this should never happen
            channelId = null;
            return;
        }
        channel.sendMessage(getBirthdayMessage(member.getUser())).queue();
    }

    /**
     * Gets the birthday message for a user
     *
     * @param user the user to get the message for
     * @return the birthday message
     */
    public String getBirthdayMessage(User user) {
        return message
                .replace("{user}", user.getAsMention())
                .replace("{age}", String.valueOf(calculateAge(user.getId())));
    }

    @Override
    public void reset() {
        birthdays.clear();
        channelId = null;
    }
}
